from pathlib import Path
from typing import Literal, overload

import yaml

from ._config_validators import (
    DatasetSimulatorConfig,
    EnsOptMDConfig,
    FlexibleFittingConfig,
    GMMFitConfig,
)


@overload
def load_config(
    path_to_config: str | Path,
    config_mode: Literal["data simulation"],
) -> DatasetSimulatorConfig: ...


@overload
def load_config(
    path_to_config: str | Path,
    config_mode: Literal["ensemble optimization"],
) -> EnsOptMDConfig: ...


@overload
def load_config(
    path_to_config: str | Path,
    config_mode: Literal["gmm fitting"],
) -> GMMFitConfig: ...
@overload
def load_config(
    path_to_config: str | Path,
    config_mode: Literal["flexible fitting"],
) -> FlexibleFittingConfig: ...


def load_config(
    path_to_config: str | Path,
    config_mode: Literal[
        "data simulation", "ensemble optimization", "gmm fitting", "flexible fitting"
    ],
) -> DatasetSimulatorConfig | EnsOptMDConfig | GMMFitConfig | FlexibleFittingConfig:
    """
    Load a configuration file and parse it into the appropriate configuration object.

    **Arguments:**
    - `path_to_config`
        Path to the configuration file (YAML format).
    - `config_mode`
        Type of configuration to load. Must be one of "data simulation",
        "ensemble optimization", "gmm fitting", or "flexible fitting".

    **Returns:**
    - Config Object
        An instance of the appropriate configuration class based on the `config_mode`.
    """
    with open(path_to_config) as f:
        config_dict = yaml.safe_load(f)

    if config_mode == "data simulation":
        return DatasetSimulatorConfig(**config_dict)
    elif config_mode == "ensemble optimization":
        return EnsOptMDConfig(**config_dict)
    elif config_mode == "gmm fitting":
        return GMMFitConfig(**config_dict)
    elif config_mode == "flexible fitting":
        return FlexibleFittingConfig(**config_dict)
    else:
        raise ValueError(
            f"Unknown config_mode: {config_mode}. Must be one of"
            + " 'data simulation', 'ensemble optimization',"
            " 'gmm fitting', or 'flexible fitting'."
        )



"""
Functions for running MD simulations using OpenMM.

Functions
---------
run_md_openmm
    Run MD simulations using OpenMM
"""

import os
import pathlib
import shutil
import warnings
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import Any
from typing_extensions import override

import jax
import jax.numpy as jnp
import mdtraj
import numpy as np
from jaxtyping import Array, Float, Int


try:
    import openmm
    import openmm.app as openmm_app
    import openmm.unit as openmm_unit

    _HAS_OPENMM = True

except ImportError:
    _HAS_OPENMM = False
    warnings.warn(
        "OpenMM is not installed. Please install OpenMM if using any features "
        + "that use molecular dynamics, e.g., ensemble optimization "
        + "or flexible fitting."
    )


from ..base_prior_projector import AbstractEnsemblePriorProjector, AbstractPriorProjector


def _get_default_md_params() -> dict:
    return {
        "forcefield": "amber14-all.xml",
        "water_model": "amber14/tip3p.xml",
        "nonbondedMethod": openmm_app.PME,
        "nonbondedCutoff": 1.0 * openmm_unit.nanometer,
        "constraints": openmm_app.HBonds,
        "temperature": 300.0 * openmm_unit.kelvin,
        "friction": 1.0 / openmm_unit.picosecond,
        "timestep": 0.002 * openmm_unit.picoseconds,
        "platform": "CPU",
        "properties": {"Threads": "1"},
    }


class SteeredMDSimulator(AbstractPriorProjector, strict=True):
    n_steps: Int
    simulation: openmm_app.Simulation
    restrain_atom_list: list[Int]
    base_state_file_path: str

    def __init__(
        self,
        path_to_initial_pdb: str | pathlib.Path,
        n_steps: Int,
        restrain_atom_list: list[Int],
        parameters_for_md: dict,
        base_state_file_path: str,
        *,
        make_simulation_fn: Callable | None = None,
    ):
        if not _HAS_OPENMM:
            raise ImportError(
                "OpenMM is not installed. Please install OpenMM if using any features "
                + "that use molecular dynamics, e.g., the ensemble optimization pipeline "
                + "or flexible fitting."
            )
        pdb = openmm_app.PDBFile(str(path_to_initial_pdb))
        self.restrain_atom_list = restrain_atom_list

        self.base_state_file_path = _validate_base_state_file_path(base_state_file_path)

        self.n_steps = n_steps
        if make_simulation_fn is None:
            parameters_for_md = _validate_and_set_params_for_md(parameters_for_md)
            self.simulation = _default_make_sim_fn(parameters_for_md, pdb.topology)

        else:
            self.simulation = make_simulation_fn(parameters_for_md, pdb.topology)

        self.simulation.context.setPositions(pdb.positions)

    @override
    def initialize(self, init_state: str | None = None) -> str:
        if init_state is not None:
            assert pathlib.Path(init_state).exists(), (
                "init_state does not exist. "
                "Please set to None or provide valid state file."
            )
            self.simulation.loadState(str(init_state))
            self.simulation.step(1000)
            path_to_state_file = f"{self.base_state_file_path}0.xml"
            if os.path.exists(path_to_state_file):
                if Path(init_state).samefile(path_to_state_file):
                    Warning(
                        "The provided init_state has the same base name as the "
                        + "base_state_file_path. "
                        + "This may cause overwriting of the state file."
                    )
                    path_to_state_file = f"{self.base_state_file_path}1.xml"

        else:
            path_to_state_file = f"{self.base_state_file_path}0.xml"
            self.simulation.minimizeEnergy()
            self.simulation.step(1000)

        self.simulation.saveState(path_to_state_file)

        return path_to_state_file

    @override
    def __call__(
        self,
        ref_walkers: Float[Array, "n_atoms 3"],
        state: str,
        bias_constant: float,
        *,
        reporter=None,
    ) -> tuple[Float[Array, "n_atoms 3"], str]:
        _assert_is_valid_state_file(state, self.base_state_file_path)

        simulation = _add_restraint_force_to_simulation(
            self.simulation,
            mdtraj.Trajectory(
                ref_walkers / 10.0, self.simulation.topology
            ).openmm_positions(0),
            self.restrain_atom_list,
            bias_constant,
        )

        # print("Reinitialize")
        simulation.context.reinitialize()

        # print("Loading state")
        simulation.loadState(state)

        # add Reporter
        if reporter is not None:
            simulation.reporters.append(reporter)

        # print("Running Simulation")
        simulation.step(self.n_steps)
        positions = simulation.context.getState(getPositions=True).getPositions()
        velocities = simulation.context.getState(getVelocities=True).getVelocities()
        # print("Cleaning up")
        simulation = _remove_last_force_from_simulation(simulation)
        simulation.context.reinitialize()  # preserveState=True)

        simulation.context.setPositions(positions)
        simulation.context.setVelocities(velocities)

        state = _get_next_state_file_path(self.base_state_file_path, state)
        simulation.saveState(state)

        # print("Saved states... Finishing.")

        positions = (
            simulation.context.getState(getPositions=True)
            .getPositions(asNumpy=True)
            .value_in_unit(openmm_unit.angstrom)
        )
        if reporter is not None:
            simulation.reporters.pop(-1)
        return jnp.array(positions), state


class EnsembleSteeredMDSimulator(AbstractEnsemblePriorProjector, strict=True):
    projectors: list[SteeredMDSimulator]

    def __init__(self, md_simulators: list[SteeredMDSimulator]):
        if not _HAS_OPENMM:
            raise ImportError(
                "OpenMM is not installed. Please install OpenMM if using any features "
                + "that use molecular dynamics, e.g., the ensemble optimization pipeline "
                + "or flexible fitting."
            )
        self.projectors = md_simulators

    @override
    def __call__(
        self,
        ref_positions: Float[Array, "n_walkers n_atoms 3"],
        states: list[str],
        bias_constant: float,
        *,
        reporters: list[Any] | None = None,
    ) -> tuple[Float[Array, "n_walkers n_atoms 3"], list[str]]:
        projected_walkers = np.zeros_like(ref_positions)
        if reporters is None:
            reporters = [None] * len(self.projectors)
        for i, projector in enumerate(self.projectors):
            projected_walkers[i], states[i] = projector(
                ref_positions[i], states[i], bias_constant, reporter=reporters[i]
            )
        return jnp.array(projected_walkers), states


def compute_biasing_constant(
    target_percentage: Float,
    path_to_initial_pdb: str,
    positions_for_bias,
    n_steps: Int,
    stride: Int = 1,
    atom_selection: str = "not element H",
    *,
    make_simulation_fn: Callable[[dict, openmm_app.Topology], openmm_app.Simulation]
    | None = None,
    parameters_for_md: dict = {},
):
    if not _HAS_OPENMM:
        raise ImportError(
            "OpenMM is not installed. Please install OpenMM if using any features "
            + "that use molecular dynamics, e.g., the ensemble optimization pipeline "
            + "or flexible fitting."
        )
    if make_simulation_fn is None:
        parameters_for_md = _validate_and_set_params_for_md(parameters_for_md)
        make_simulation_fn = _default_make_sim_fn

    restrain_atom_list = mdtraj.load(str(path_to_initial_pdb)).topology.select(
        atom_selection
    )

    pdb = openmm_app.PDBFile(str(path_to_initial_pdb))
    simulation = make_simulation_fn(parameters_for_md, pdb.topology)
    simulation.context.setPositions(pdb.positions)
    simulation.minimizeEnergy()

    # initial_positions = simulation.context.getState(
    #     getPositions=True
    # ).getPositions()

    # initial_positions = mdtraj.load(str(path_to_initial_pdb)).openmm_positions(0)

    dir_exists = pathlib.Path("./tmp_biasing_comp").exists()
    os.makedirs("./tmp_biasing_comp", exist_ok=True)
    path_to_traj = "./tmp_biasing_comp/traj_for_force.xtc"
    simulation.reporters.append(
        openmm_app.XTCReporter(path_to_traj, reportInterval=stride)
    )
    simulation.step(n_steps)

    traj = mdtraj.load(path_to_traj, top=path_to_initial_pdb)

    simulation = make_simulation_fn(parameters_for_md, pdb.topology)
    md_forces = _compute_regular_force(traj, simulation)

    simulation = make_simulation_fn(parameters_for_md, pdb.topology)
    bias_forces = _compute_biasing_force(
        traj, simulation, restrain_atom_list, positions_for_bias
    )

    k_value = _compute_k_value(
        jnp.array(md_forces),
        jnp.array(bias_forces),
        target_percentage,
        restrain_atom_list,
    ).mean()

    os.remove(path_to_traj)
    if not dir_exists:
        shutil.rmtree("./tmp_biasing_comp")

    return k_value


@partial(jax.vmap, in_axes=(0, 0, None, None))
def _compute_k_value(base_force, bias_force, target_percentage, restrain_atom_list):
    force1 = base_force[restrain_atom_list, :].flatten()
    force2 = bias_force[restrain_atom_list, :].flatten()

    rho = jnp.dot(force1, force2) / jnp.sum(force2**2)
    R = jnp.sum(force1**2) / jnp.sum(force2**2)

    a = 1.0 - target_percentage**2
    b = -2.0 * target_percentage**2 * rho
    c = -R * target_percentage**2

    return -b + jnp.sqrt(b**2 - 4.0 * a * c) / (2.0 * a)


def _compute_regular_force(trajectory: mdtraj.Trajectory, simulation):
    forces = np.zeros((trajectory.n_frames, trajectory.n_atoms, 3))
    for i in range(trajectory.n_frames):
        simulation.context.setPositions(trajectory.openmm_positions(i))
        forces[i] = np.array(
            simulation.context.getState(getForces=True).getForces(asNumpy=True)
        )
    return forces


def _compute_biasing_force(
    trajectory: mdtraj.Trajectory,
    simulation,
    restrain_atom_list,
    bias_positions,
):
    RMSD_value = openmm.RMSDForce(
        bias_positions,
        restrain_atom_list,
    )

    force_RMSD = openmm.CustomCVForce("0.5 * k * RMSD^2")
    force_RMSD.addGlobalParameter("k", 1.0)
    force_RMSD.addCollectiveVariable("RMSD", RMSD_value)
    simulation.system.addForce(force_RMSD)

    n_forces = len(simulation.system.getForces())
    for i in range(n_forces - 1):
        simulation.system.removeForce(0)

    simulation.context.reinitialize()
    forces = np.zeros((trajectory.n_frames, trajectory.n_atoms, 3))
    for i in range(trajectory.n_frames):
        simulation.context.setPositions(trajectory.openmm_positions(i))
        forces[i] = np.array(
            simulation.context.getState(getForces=True).getForces(asNumpy=True)
        )
    return forces


def _validate_base_state_file_path(base_state_file_path: str) -> str:
    # check if the path exists
    base_dir = os.path.dirname(base_state_file_path)
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    if not base_state_file_path.endswith("_"):
        return f"{base_state_file_path}_it"
    else:
        return f"{base_state_file_path}it"


def _get_next_state_file_path(
    base_state_file_path: str,
    curr_state_file: str,
) -> str:
    # get the run counter from the last file
    last_counter = int(curr_state_file.split(base_state_file_path)[-1].split(".xml")[0])
    return f"{base_state_file_path}{last_counter + 1}.xml"


def _assert_is_valid_state_file(
    state_file: str,
    base_state_file_path: str,
) -> None:
    assert base_state_file_path in state_file, (
        "State file does not match base state file path. "
        "Please provide a valid state file."
    )

    counter = state_file.split(base_state_file_path)[-1].split(".xml")[0]
    try:
        int(counter)
    except ValueError:
        raise ValueError(
            f"State file should be formatted as base_state_file_path + <int>.xml. "
            f"Got {state_file} instead."
        )
    return


def _add_restraint_force_to_simulation(
    simulation: openmm_app.Simulation,
    positions: openmm_unit.Quantity,
    restrain_atom_list: list[int],
    bias_constant_in_kj_per_mol_angs: float,
) -> openmm_app.Simulation:
    RMSD_value = openmm.RMSDForce(
        positions,
        restrain_atom_list,
    )

    force_RMSD = openmm.CustomCVForce("0.5 * k * RMSD^2")
    force_RMSD.addGlobalParameter("k", bias_constant_in_kj_per_mol_angs)
    force_RMSD.addCollectiveVariable("RMSD", RMSD_value)
    simulation.system.addForce(force_RMSD)

    return simulation


def _remove_last_force_from_simulation(
    simulation: openmm_app.Simulation,
) -> openmm_app.Simulation:
    n_forces = len(simulation.system.getForces())
    simulation.system.removeForce(n_forces - 1)
    return simulation


def _default_make_sim_fn(parameters_for_md: dict, topology) -> openmm_app.Simulation:
    forcefield = _create_forcefield(parameters_for_md)
    integrator = _create_integrator(parameters_for_md)
    platform = _create_platform(parameters_for_md)
    system = _create_system(parameters_for_md, forcefield, topology)

    simulation = openmm_app.Simulation(
        topology,
        system,
        integrator,
        platform,
        parameters_for_md["properties"],
    )

    return simulation


def _create_forcefield(parameters_for_md: dict) -> openmm_app.ForceField:
    return openmm_app.ForceField(
        parameters_for_md["forcefield"], parameters_for_md["water_model"]
    )


def _create_integrator(parameters_for_md: dict) -> openmm.Integrator:
    return openmm.LangevinIntegrator(
        parameters_for_md["temperature"],
        parameters_for_md["friction"],
        parameters_for_md["timestep"],
    )


def _create_system(
    parameters_for_md: dict,
    forcefield: openmm_app.ForceField,
    topology: openmm_app.Topology,
) -> openmm.System:
    system = forcefield.createSystem(
        topology,
        nonbondedMethod=parameters_for_md["nonbondedMethod"],
        nonbondedCutoff=parameters_for_md["nonbondedCutoff"],
        constraints=parameters_for_md["constraints"],
    )

    return system


def _create_platform(parameters_for_md: dict) -> openmm.Platform:
    return openmm.Platform.getPlatformByName(parameters_for_md["platform"])


def _validate_and_set_params_for_md(
    parameters_for_md: dict,
) -> dict:
    default_md_params = _get_default_md_params()
    assert set(parameters_for_md.keys()).issubset(default_md_params)
    for key, value in default_md_params.items():
        if key not in parameters_for_md:
            parameters_for_md[key] = value

    return parameters_for_md



"""
Weight and position optimizers for ensemble refinement.
"""

from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
import jax_dataloader as jdl
from jaxopt import ProjectedGradient
from jaxopt.projection import projection_simplex
from jaxtyping import Array, Float, Int

from .._pose_search import HierarchicalSO3GridSearch
from . import ImagesToEnsembleLikelihoodFn
from .base_optimizer import AbstractEnsembleParameterOptimizer


class ProjGradDescWeightOptimizer(AbstractEnsembleParameterOptimizer):
    n_steps: Int
    ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn
    pose_search: HierarchicalSO3GridSearch | None

    def __init__(
        self,
        n_steps: Int,
        ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn,
        pose_search: HierarchicalSO3GridSearch | None = None,
    ):
        self.n_steps = n_steps
        self.ensemble_likelihood_fn = ensemble_likelihood_fn
        self.pose_search = pose_search

    @override
    def __call__(
        self,
        walkers: Float[Array, "n_walkers n_atoms 3"],
        weights: Float[Array, " n_walkers"],
        dataloader: jdl.DataLoader,
    ):
        """
        Optimize the weights of the walkers using projected gradient descent
        using all images.

        **Arguments:**
            walkers: The current positions of the walkers.
            weights: The current weights of the walkers.
            dataloader: The dataloader for the data.
            args: Additional arguments for the likelihood function.
            This should be a tuple with the following elements:
                - `amplitudes`: The Gaussian amplitudes for each atom.
                - `variances`: The Gaussian variances for each atom.
                - `noise_variance`: The noise variance for the data. If None, the
                    noise variance is marginalized.

        **Returns:**
            The optimized weights of the walkers.
        """
        likelihood_matrix = (
            self.ensemble_likelihood_fn.compute_full_log_likelihood_matrix(
                walkers, dataloader, prints_progress=True
            )
        )
        return self.optimize_with_precomputed_likelihood_matrix(
            likelihood_matrix, weights
        )

    def optimize_with_precomputed_likelihood_matrix(
        self,
        likelihood_matrix: Float[Array, "n_images n_walkers"],
        init_weights: None | Float[Array, " n_walkers"] = None,
    ):
        """
        Optimize the weights of the walkers using projected gradient descent
        using a precomputed likelihood matrix.

        **Arguments:**
            weights: The current weights of the walkers.
            likelihood_matrix: The precomputed likelihood matrix for each image
                and walker.

        **Returns:**
            The optimized weights of the walkers.
        """
        if init_weights is None:
            init_weights = (
                jnp.ones(likelihood_matrix.shape[1]) / likelihood_matrix.shape[1]
            )
        weights = _optimize_weights(
            init_weights, likelihood_matrix, self.ensemble_likelihood_fn, self.n_steps
        )
        return weights


@eqx.filter_jit
def _optimize_weights(
    weights: Float[Array, " n_walkers"],
    log_likelihood_matrix: Float[Array, "n_images n_walkers"],
    ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn,
    n_steps: Int = 500,
) -> Float[Array, " n_walkers"]:
    def loss_fn(w, llm):
        return -ensemble_likelihood_fn.compute_from_log_likelihood_matrix(llm, w)

    pg = ProjectedGradient(
        fun=loss_fn,
        projection=projection_simplex,
        maxiter=n_steps,
    )
    return pg.run(weights, llm=log_likelihood_matrix).params



import logging
import os
import pathlib
from typing import Any

import equinox as eqx
import jax.numpy as jnp
import mdtraj
import numpy as np
import optax
from jaxtyping import Array, Float, Int
from mdtraj.formats import XTCTrajectoryFile
from tqdm import tqdm

from cryojax_eo.ensemble_optimization import (
    SteeredMDSimulator,
)
from cryojax_eo.utils import ModelToVolumeAligner, rigid_align_positions

from .._cross_corelation.optimizer import SteepestDescWalkerFlexibleFitting


class FlexibleFittingPipeline(eqx.Module):
    """
    Ensemble refinement pipeline using OpenMM for molecular dynamics simulation.
    """

    prior_projector: SteeredMDSimulator
    walker_optimizer: SteepestDescWalkerFlexibleFitting
    n_steps: int
    prealigned_structure: mdtraj.Trajectory
    model_to_volume_aligner: ModelToVolumeAligner | None
    atom_indices_for_opt: Int[Array, " n_atoms_for_opt"]

    def __init__(
        self,
        prior_projector: SteeredMDSimulator,
        walker_optimizer: SteepestDescWalkerFlexibleFitting,
        n_steps: int,
        prealigned_structure: mdtraj.Trajectory,
        atom_indices_for_opt: Int[Array, " n_atoms_for_opt"],
        model_to_volume_aligner: ModelToVolumeAligner | None = None,
    ):
        assert n_steps > 0, "n_steps must be positive"
        assert atom_indices_for_opt.ndim == 1, "atom_indices_for_opt must be a 1D array."
        assert (
            len(atom_indices_for_opt) > 0
        ), "atom_indices_for_opt must contain at least one index."

        self.prior_projector = prior_projector
        self.walker_optimizer = walker_optimizer
        self.n_steps = n_steps
        self.prealigned_structure = prealigned_structure
        self.atom_indices_for_opt = atom_indices_for_opt
        self.model_to_volume_aligner = model_to_volume_aligner

    def run(
        self,
        initial_walker: Float[Array, "n_atoms 3"],
        reference_volume: Float[Array, "n_pixels n_pixels n_pixels"],
        bias_constant_scheduler: optax.ScalarOrSchedule,
        *,
        output_directory: str | pathlib.Path,
        initial_state_for_projector: Any = None,
    ) -> tuple[Float[Array, "n_atoms 3"], Any]:
        logging.info("Initializing projector...")
        md_state = self.prior_projector.initialize(initial_state_for_projector)
        logging.info("Projector initialized.")

        unit_cell_vectors = np.array(
            self.prior_projector.simulation.topology.getPeriodicBoxVectors()._value
        )

        ref_positions = (
            jnp.asarray(self.prealigned_structure.xyz[0]) * 10.0
        )  # Convert from nm to Angstroms
        walker = initial_walker.copy()

        logging.info("Preparing writers for output...")
        writer = XTCTrajectoryFile(os.path.join(output_directory, "traj_walker.xtc"), "w")
        logging.info("Writers prepared.")

        logging.info("Aligning walker to reference structure...")
        walker = _align_walker_to_reference(
            walker, ref_positions, self.atom_indices_for_opt
        )
        logging.info("Walkers aligned.")

        progress_bar = tqdm(range(self.n_steps), desc="Flexible Fitting", leave=True)
        for i in progress_bar:
            """
            if stride_for_pose is True:
                new_dataset = pose_estimation(walker)
                dataloader = create_dataloader...
            """

            logging.info("Likelihood Optimization: ")
            loss, tmp_walker = self.walker_optimizer(
                walker[self.atom_indices_for_opt, :],
                reference_volume,
            )

            ref_walker = walker.at[self.atom_indices_for_opt, :].set(tmp_walker)
            ref_walker.block_until_ready()
            logging.info("Likelihood Optimization done.")

            logging.info("Prior Projection: ")
            walker, md_state = self.prior_projector(
                ref_walker, md_state, bias_constant_scheduler(i)
            )

            logging.info("Aligning walker to reference structure...")
            walker = _align_walker_to_reference(
                walker, ref_positions, self.atom_indices_for_opt
            )
            logging.info("Walker aligned.")

            if self.model_to_volume_aligner is not None:
                logging.info("   Aligning walker to volume...")
                walker = _align_walkers_to_volume(
                    walker,
                    self.model_to_volume_aligner,
                    self.atom_indices_for_opt,
                    self.walker_optimizer.model_to_vol_loss_fn.amplitudes,
                    self.walker_optimizer.model_to_vol_loss_fn.variances,
                )
                logging.info("   Walkers aligned to volume.")

            logging.info("Write trajectory to files...")
            writer.write(walker / 10.0)
            _write_walker_to_pdb(
                walker,
                os.path.join(output_directory, "curr_walker.pdb"),
                self.prealigned_structure.topology,
                unit_cell_vectors,
            )

            progress_bar.set_description(f"Flexible Fitting (C.C: {1 - loss:.4f})")

        writer.close()
        _write_walker_to_pdb(
            walker,
            os.path.join(output_directory, "final_walker.pdb"),
            self.prealigned_structure.topology,
            unit_cell_vectors,
        )

        return walker, md_state


def _write_walker_to_pdb(walker, filename, topology, unit_cell_vectors):
    walker_as_traj = mdtraj.Trajectory(
        xyz=walker / 10.0,
        topology=topology,
    )
    walker_as_traj.unitcell_vectors = unit_cell_vectors[None, ...]
    walker_as_traj.save_pdb(filename)
    return


@eqx.filter_jit
def _align_walker_to_reference(
    walker: Float[Array, "n_atoms 3"],
    ref_positions: Float[Array, "n_atoms 3"],
    atom_indices: Int[Array, " n_atoms_for_opt"],
) -> Float[Array, "n_atoms 3"]:
    """
    Align the walker to the reference structure.
    """

    _, rot_matrix, displacement = rigid_align_positions(
        walker[atom_indices], ref_positions[atom_indices]
    )
    aligned_walker = walker @ rot_matrix.T + displacement

    return jnp.asarray(aligned_walker)


def _align_walkers_to_volume(
    walker: Float[Array, "n_atoms 3"],
    model_to_volume_aligner: ModelToVolumeAligner,
    atom_indices: Int[Array, " n_atoms_for_opt"],
    amplitudes,
    variances,
) -> Float[Array, "n_atoms 3"]:
    """
    Align the walker to the volume using the ModelToVolumeAligner.
    """
    _, solution = model_to_volume_aligner.align(
        walker[atom_indices, :],
        amplitudes,
        variances,
    )

    return walker @ solution.rotation_matrix + solution.offset



#!/usr/bin/env python3
import argparse
import datetime
import logging
import os
from pathlib import Path

import jax.numpy as jnp
import mdtraj
import numpy as np
import optax
import yaml
from cryojax.io import read_array_from_mrc
from cryojax.ndimage import fourier_crop_to_shape

from cryojax_eo.ensemble_optimization import SteeredMDSimulator
from cryojax_eo.flexible_fitting import (
    FlexibleFittingPipeline,
    ModelToVolumeCorrelationLossFn,
    SteepestDescWalkerFlexibleFitting,
)
from cryojax_eo.internal import FlexibleFittingConfig
from cryojax_eo.io import read_walkers_from_pdbs
from cryojax_eo.utils import ModelToVolumeAligner


def add_args(parser):
    parser.add_argument(
        "--config", type=str, help="Path to the config (yaml) file", required=True
    )
    return parser


def mkbasedir(out):
    if not os.path.exists(out):
        try:
            os.makedirs(out)
        except (FileExistsError, PermissionError):
            raise ValueError("Output path does not exist and cannot be created.")
    return


def warnexists(out):
    if os.path.exists(out):
        Warning(f"Warning: {out} already exists. Overwriting.")
    return


def _make_atom_list(atom_selection, topology) -> np.ndarray:
    suffix = Path(atom_selection).suffix
    if suffix in [".txt", ".npy"]:
        atom_list = np.loadtxt(atom_selection, dtype=int)
    else:
        atom_list = topology.select(atom_selection)
    return np.array(atom_list)


def run_flexible_fitting(flexible_fitting_config: FlexibleFittingConfig):
    config = dict(flexible_fitting_config.model_dump())

    # Load the initial walkers and reference structure

    logging.debug("Loading atomic models...")
    initial_walker, variances, amplitudes = read_walkers_from_pdbs(
        [config["path_to_atomic_model"]],
        loads_b_factors=config["loads_b_factors"],
    )

    ref_structure = mdtraj.load(str(config["path_to_prealigned_atomic_model"]))
    ref_structure = ref_structure.center_coordinates(mass_weighted=True)

    atom_list = _make_atom_list(config["atom_selection"], ref_structure.topology)
    initial_walker = initial_walker[0]
    variances = variances[atom_list]
    amplitudes = amplitudes[atom_list]

    logging.debug("Atomic model loaded.")

    logging.debug("Loading reference volume...")
    reference_volume, voxel_size = read_array_from_mrc(
        config["reference_volume_params"]["path_to_reference_volume"],
        loads_grid_spacing=True,
    )
    reference_volume = jnp.asarray(reference_volume)

    if config["reference_volume_params"]["reference_volume_voxel_size"] is not None:
        voxel_size = config["reference_volume_params"]["reference_volume_voxel_size"]

    box_size_align = int(config["reference_volume_params"]["rigid_alignment_box_size"])
    voxel_size_ds = voxel_size * reference_volume.shape[0] / box_size_align

    model_aligner = ModelToVolumeAligner(
        fourier_crop_to_shape(
            reference_volume, (box_size_align, box_size_align, box_size_align)
        ),
        voxel_size=voxel_size_ds,
    )

    box_size_ff = int(config["reference_volume_params"]["flexible_fitting_box_size"])
    voxel_size_ff = voxel_size * reference_volume.shape[0] / box_size_ff
    reference_volume = fourier_crop_to_shape(
        reference_volume, (box_size_ff, box_size_ff, box_size_ff)
    )
    logging.debug("Reference volume loaded.")

    # Construct prior projector
    prior_projector = SteeredMDSimulator(
        path_to_initial_pdb=config["path_to_atomic_model"],
        n_steps=config["projector_params"]["n_steps"],
        restrain_atom_list=atom_list.tolist(),
        parameters_for_md={
            "platform": config["projector_params"]["platform"],
            "properties": config["projector_params"]["platform_properties"],
        },
        base_state_file_path=os.path.join(config["path_to_output"], "states_proj/state_"),
    )

    # Construct likelihood optimizer
    model_to_vol_loss_fn = ModelToVolumeCorrelationLossFn(
        amplitudes=amplitudes,
        variances=variances,
        voxel_size=voxel_size_ff,
        volume_shape=(box_size_ff, box_size_ff, box_size_ff),
        batch_size_for_z_planes=config["walker_optimizer_params"][
            "batch_size_for_z_planes"
        ],
        n_batches_of_atoms=config["walker_optimizer_params"]["n_batches_of_atoms"],
    )

    walker_optimizer = SteepestDescWalkerFlexibleFitting(
        n_steps=config["walker_optimizer_params"]["n_steps"],
        step_size=config["walker_optimizer_params"]["step_size"],
        model_to_vol_loss_fn=model_to_vol_loss_fn,
    )

    # Construct the ensemble optimization pipeline
    flexible_fitting_pipeline = FlexibleFittingPipeline(
        prior_projector=prior_projector,
        walker_optimizer=walker_optimizer,
        n_steps=config["n_steps"],
        prealigned_structure=ref_structure,
        atom_indices_for_opt=atom_list,
        model_to_volume_aligner=model_aligner,
    )

    # Running the optimization

    if isinstance(config["projector_params"]["bias_constant_in_kjpermol"], float):
        bias_constant_scheduler = optax.constant_schedule(
            config["projector_params"]["bias_constant_in_kjpermol"]
        )
    elif (
        isinstance(config["projector_params"]["bias_constant_in_kjpermol"], list)
        and len(config["projector_params"]["bias_constant_in_kjpermol"]) == 2
    ):
        bias_constant_scheduler = optax.linear_schedule(
            init_value=config["projector_params"]["bias_constant_in_kjpermol"][0],
            end_value=config["projector_params"]["bias_constant_in_kjpermol"][1],
            transition_steps=config["n_steps"],
        )
    else:
        raise ValueError(
            "bias_constant_in_kjpermol must be a float or a list of two floats."
        )

    final_walker, md_state = flexible_fitting_pipeline.run(
        initial_walker=initial_walker,
        reference_volume=reference_volume,
        bias_constant_scheduler=bias_constant_scheduler,
        output_directory=config["path_to_output"],
        initial_state_for_projector=config["projector_params"]["path_to_initial_state"],
    )

    jnp.save(
        os.path.join(config["path_to_output"], "final_walker.npy"),
        final_walker,
    )

    return final_walker, md_state


def main(args):
    with open(args.config) as f:
        config_dict = yaml.safe_load(f)
        config = FlexibleFittingConfig(**config_dict)

    warnexists(config.path_to_output)
    mkbasedir(config.path_to_output)

    logger = logging.getLogger()
    logger.handlers.clear()

    logger_fname = datetime.datetime.now().strftime("%Y-%m-%d")
    logger_fname = os.path.join(config.path_to_output, logger_fname + ".log")

    fhandler = logging.FileHandler(filename=logger_fname, mode="a")
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    fhandler.setFormatter(formatter)
    logger.addHandler(fhandler)
    logger.setLevel(logging.INFO)

    config_fname = os.path.basename(args.config)
    with open(os.path.join(config.path_to_output, config_fname), "w") as f:
        yaml.dump(config_dict, f, default_flow_style=False)

    logging.info(
        f"A copy of the used config file has been written "
        f"to {os.path.join(config.path_to_output, config_fname)}"
    )

    logging.info("Running ensemble optimization...")
    run_flexible_fitting(config)
    logging.info("Ensemble optimization complete.")

    return


def main_cli():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=yaml.dump(FlexibleFittingConfig.model_json_schema(), indent=4),
    )
    main(add_args(parser).parse_args())


if __name__ == "__main__":
    main_cli()



from pathlib import Path
from typing import Literal

import mdtraj
from pydantic import (
    BaseModel,
    Field,
    FilePath,
    PositiveFloat,
    PositiveInt,
    field_validator,
    model_validator,
)

from .utils import _validate_file_with_type, _validate_files_with_type


class FFOptimizationConfig(BaseModel, extra="forbid"):
    n_steps: PositiveInt = Field(
        default=1, description="Number of steps for the optimization process."
    )
    step_size: PositiveFloat = Field(
        description="Step size in Angstroms for the optimization process."
    )

    batch_size_for_z_planes: PositiveInt = Field(
        default=1,
        description="The number of z-planes to evaluate in parallel with"
        " `jax.vmap`. By default, `1`.",
    )
    n_batches_of_atoms: PositiveInt = Field(
        default=1,
        description="The number of iterations used to evaluate the volume, "
        "where the iteration is taken over groups of atoms. "
        "This is useful if `batch_size = 1` and GPU memory is exhausted. "
        "By default, `1`.",
    )


class FFProjectorConfig(BaseModel, extra="forbid"):
    n_steps: PositiveInt = Field(
        description="Number of steps for the MD sampler. Must be greater than 0."
    )
    bias_constant_in_kjpermol: PositiveFloat | list[PositiveFloat] = Field(
        description="Biasing constant for the projection step. "
        + "Can be a single value or a list of two values for linear scheduling."
    )
    platform: Literal["CPU", "CUDA", "OpenCL"] = Field(
        default="CPU",
        description="Platform to use for the MD sampler. "
        + "Must be 'CPU', 'CUDA', or 'OpenCL'.",
    )
    platform_properties: dict[str, str] | None = Field(
        default=None,
        description="Platform properties for OpenMM. "
        + "For CPU the default is {'Threads': '1'}"
        + " and for CUDA the default is {'DeviceIndex': '0'}.",
    )

    path_to_initial_state: FilePath | None = Field(
        default=None,
        description="Path to the initial openMM state. "
        + "If None, the projector will be initialized with a random state. ",
    )

    @model_validator(mode="after")
    def validate_config(self):
        if self.platform == "CPU":
            if self.platform_properties is None:
                self.platform_properties = {"Threads": "1"}
        elif self.platform == "CUDA":
            if self.platform_properties is None:
                self.platform_properties = {"DeviceIndex": "0"}
        else:
            if self.platform_properties is None:
                raise ValueError(
                    "platform_properties must be provided for OpenCL platform"
                )
        return self

    @field_validator("path_to_initial_state")
    @classmethod
    def validate_path_to_initial_state(cls, v):
        if v is None:
            return v
        else:
            return _validate_files_with_type([v], file_types=[".xml"])[0]


class RefVolFFConfig(BaseModel, extra="forbid"):
    path_to_reference_volume: FilePath | None = Field(
        default=None,
        description="Path to the consensus volume. "
        + "Used for rigid body alignment of walkers. "
        + "If None, no alignment will be performed.",
    )

    flexible_fitting_box_size: PositiveInt = Field(
        default=128,
        description="Box size to crop the volume for flexible fitting. "
        + "The box size must be a positive integer.",
    )

    rigid_alignment_box_size: PositiveInt = Field(
        default=32,
        description="Box size to downsample the volume for alignment. "
        + "The box size must be a positive integer.",
    )

    reference_volume_voxel_size: PositiveFloat | None = Field(
        default=None,
        description="Overrides the voxel size stored in the MRC header "
        + "of the consensus volume."
        + " If None, the voxel size from the MRC header will be used.",
    )

    @field_validator("path_to_reference_volume")
    @classmethod
    def validate_path_to_reference_volume(cls, v):
        if v is not None:
            return _validate_file_with_type(v, file_type=".mrc")


class FlexibleFittingConfig(BaseModel, extra="forbid"):
    # I/O
    path_to_atomic_model: FilePath = Field(
        description="Path to the atomic model used for initialization."
        " Either in .pdb or .cif format."
    )
    path_to_prealigned_atomic_model: FilePath = Field(
        description="Path to the reference model. "
        + "This model should be aligned to the cryo-EM particles, "
        + " and will be used for alignment of the walkers during optimization."
    )

    path_to_output: Path = Field(
        description="Path to the output directory. "
        + "If it does not exist, it will be created.",
    )
    atom_selection: str | FilePath = Field(
        default="all",
        description="Selection string for atom selection, "
        + "or a txt/npy file containing atom indices",
    )

    loads_b_factors: bool = Field(
        default=False,
        description="Whether to load the thermal b-factors from the PDB file. "
        + "Only used if the atomic model is in PDB format. "
        + "Otherwise it will be ignored."
        + "Also known as Debye-Waller factors.",
    )

    # Reference Volume Params
    reference_volume_params: dict = Field(
        description="Parameters for the reference volume. "
    )

    # Pipeline
    projector_params: dict = Field(
        description="Parameters for the physics-based ensemble projector. "
        + "This is a dictionary formatted by the `FFConfigProjector` class."
    )
    walker_optimizer_params: dict = Field(
        description="Parameters for the ensemble optimizer. "
        + "This is a dictionary formatted by "
        + "the `FFOptimizationConfig` class."
    )
    # Optimization
    n_steps: PositiveInt = Field(
        description="Number of steps of cryoJAX ensemble refinement to run."
    )

    @field_validator("path_to_atomic_model")
    @classmethod
    def validate_path_to_atomic_model(cls, v):
        return _validate_files_with_type([v], file_types=[".pdb", ".cif"])[0]

    @field_validator("walker_optimizer_params")
    @classmethod
    def validate_ensemble_opt_config(cls, values):
        return dict(FFOptimizationConfig(**values).model_dump())

    @field_validator("projector_params")
    @classmethod
    def validate_md_sampler_config(cls, values):
        return dict(FFProjectorConfig(**values).model_dump())

    @field_validator("reference_volume_params")
    @classmethod
    def validate_ref_vol_config(cls, values):
        return dict(RefVolFFConfig(**values).model_dump())

    @field_validator("atom_selection")
    @classmethod
    def validate_atom_selection(cls, values):
        suffix = Path(values).suffix
        if suffix in [".txt", ".npy"]:
            assert Path(values).exists(), f"Indices File: {values} does not exist."

        elif suffix not in [".txt", ".npy", ""]:
            raise ValueError("Invalid file type for atom selection.")

        else:
            try:
                mdtraj.Topology().select(values)
            except Exception as e:
                raise ValueError(f"Invalid atom selection string: {values}. Error: {e}")
        return values



from ._run_ensemble_optimization import (
    run_ensemble_optimization_with_md as run_ensemble_optimization_with_md,
)
from ._run_flexible_fitting import (
    run_flexible_fitting as run_flexible_fitting,
)



from typing import Any, cast

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
from cryojax.jax_util import filter_bmap
from jax_dataloader import DataLoader
from jaxtyping import Array, Float
from tqdm import tqdm

from .single_likelihood import AbstractImageToWalkerLogLikelihoodFn


class AbstractImagesToEnsembleLikelihoodFn(eqx.Module):
    image_to_walker_likelihood_fn: eqx.AbstractVar[AbstractImageToWalkerLogLikelihoodFn]
    n_walkers_in_parallel: eqx.AbstractVar[int]
    n_images_in_parallel: eqx.AbstractVar[int]

    def compute_log_likelihood_matrix(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        images: Float[Array, "n_images y x"],
        image_config: cxs.BasicImageConfig,
        poses_per_walker: cxs.AbstractPose,
        transfer_theories: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ) -> Float[Array, "n_images n_walkers"]:
        raise NotImplementedError

    def __call__(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        weights: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        images: Float[Array, "n_images y x"],
        image_config: cxs.BasicImageConfig,
        poses_per_walker: cxs.AbstractPose,
        transfer_theory: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ):
        raise NotImplementedError


class ImagesToEnsembleLikelihoodFn(AbstractImagesToEnsembleLikelihoodFn):
    image_to_walker_likelihood_fn: AbstractImageToWalkerLogLikelihoodFn
    n_walkers_in_parallel: int = 1
    n_images_in_parallel: int = 1

    def compute_log_likelihood_matrix(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        images: Float[Array, "n_images y x"],
        image_config: cxs.BasicImageConfig,
        poses_per_walker: cxs.AbstractPose,
        transfer_theories: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ) -> Float[Array, "n_images n_walkers"]:
        return _compute_likelihoods_over_walkers(
            self.image_to_walker_likelihood_fn,
            walkers,
            images,
            image_config,
            poses_per_walker,
            transfer_theories,
            per_particle_args,
            self.n_walkers_in_parallel,
            self.n_images_in_parallel,
        ).T

    def compute_from_log_likelihood_matrix(
        self,
        log_likelihood_matrix: Float[Array, "n_images n_walkers"],
        weights: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
    ):
        return jnp.mean(
            jax.scipy.special.logsumexp(
                a=log_likelihood_matrix, b=weights[None, :], axis=1
            )
        )

    def compute_full_log_likelihood_matrix(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        dataloader: DataLoader,
        *,
        prints_progress=False,
    ) -> Float[Array, "n_images n_walkers"]:
        compute_likelihood_matrix_fn = eqx.filter_jit(self.compute_log_likelihood_matrix)

        shuffle = dataloader.dataloader.shuffle  # save the original shuffle state
        dataloader.dataloader.shuffle = False
        # Compute the likelihood matrix for each batch in the dataloader
        likelihood_matrix = []

        iterable = (
            dataloader
            if not prints_progress
            else tqdm(dataloader, desc="Computing full likelihood matrix", unit=" batch")
        )
        for batch in iterable:
            poses_per_walker = jax.tree.map(
                lambda x: jnp.repeat(x[None, :], repeats=walkers.shape[0], axis=0),
                batch["particle_stack"]["parameters"]["pose"],
            )
            poses_per_walker = cast(cxs.EulerAnglePose, poses_per_walker)

            lklhood_matrix = compute_likelihood_matrix_fn(
                walkers,
                batch["particle_stack"]["images"],
                batch["particle_stack"]["parameters"]["image_config"],
                poses_per_walker,
                batch["particle_stack"]["parameters"]["transfer_theory"],
                batch["per_particle_args"],
            )
            likelihood_matrix.append(lklhood_matrix)

        # restore the original shuffle state
        dataloader.dataloader.shuffle = shuffle

        # Concatenate the likelihood matrices from all batches
        return jnp.concatenate(likelihood_matrix, axis=0)

    def __call__(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        weights: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        images: Float[Array, "n_images y x"],
        image_config: cxs.BasicImageConfig,
        poses_per_walker: cxs.AbstractPose,
        transfer_theories: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ):
        likelihood_matrix = self.compute_log_likelihood_matrix(
            walkers,
            images,
            image_config,
            poses_per_walker,
            transfer_theories,
            per_particle_args,
        )
        return self.compute_from_log_likelihood_matrix(likelihood_matrix, weights)


def _compute_likelihoods_single_walker_bmap(
    likelihood_fn: AbstractImageToWalkerLogLikelihoodFn,
    walker: Float[Array, "n_atoms n_gaussians_per_atom"],
    images: Float[Array, "n_images y x"],
    image_config: cxs.BasicImageConfig,
    poses_per_walker: cxs.AbstractPose,
    transfer_theories: cxs.ContrastTransferTheory,
    per_particle_args: Any,
    n_images_in_parallel: int,
):
    @eqx.filter_vmap(in_axes=(0, None, 0, 0, 0))
    def _body_fn(img, ic, p, tf, ppa):
        return likelihood_fn(
            walker,
            img,
            ic,
            p,
            tf,
            ppa,
        )

    return filter_bmap(
        lambda x: _body_fn(x[0], image_config, x[1], x[2], x[3]),
        (images, poses_per_walker, transfer_theories, per_particle_args),
        batch_size=n_images_in_parallel,
    )


def _compute_likelihoods_over_walkers(
    likelihood_fn: AbstractImageToWalkerLogLikelihoodFn,
    walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
    images: Float[Array, "n_images y x"],
    image_config: cxs.BasicImageConfig,
    poses_per_walker: cxs.AbstractPose,
    transfer_theories: cxs.ContrastTransferTheory,
    per_particle_args: Any,
    n_walkers_in_parallel: int,
    n_images_in_parallel: int,
):
    @eqx.filter_vmap(in_axes=(0, None, None, 0, None, None))
    def _body_fn(w, im, ic, p, tf, ppa):
        return _compute_likelihoods_single_walker_bmap(
            likelihood_fn,
            w,
            im,
            ic,
            p,
            tf,
            ppa,
            n_images_in_parallel,
        )

    return filter_bmap(
        lambda x: _body_fn(
            x[0], images, image_config, x[1], transfer_theories, per_particle_args
        ),
        xs=(walkers, poses_per_walker),
        batch_size=n_walkers_in_parallel,
    )



from ._cross_corelation import (
    ModelToVolumeCorrelationLossFn as ModelToVolumeCorrelationLossFn,
    SteepestDescWalkerFlexibleFitting as SteepestDescWalkerFlexibleFitting,
)
from ._pipeline import FlexibleFittingPipeline as FlexibleFittingPipeline



from .model_to_volume_loss import (
    ModelToVolumeCorrelationLossFn as ModelToVolumeCorrelationLossFn,
)
from .optimizer import (
    SteepestDescWalkerFlexibleFitting as SteepestDescWalkerFlexibleFitting,
)
